@@ -1,6 +1,7 @@
 import asyncio
 import getpass
 import logging
+import os
 import sys
 from contextlib import contextmanager
 from datetime import datetime
@@ -17,9 +18,11 @@ from pyhanko.config import (
     CLIConfig,
     LogConfig,
     PemDerSignatureConfig,
+    PKCS11PinEntryMode,
     PKCS11SignatureConfig,
     PKCS12SignatureConfig,
     StdLogOutput,
+    TokenCriteria,
     init_validation_context_kwargs,
     parse_cli_config,
     parse_logging_config,
@@ -28,6 +31,7 @@ from pyhanko.pdf_utils import crypt, misc
 from pyhanko.pdf_utils.config_utils import ConfigurationError
 from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
 from pyhanko.pdf_utils.layout import LayoutError
+from pyhanko.pdf_utils.misc import isoparse
 from pyhanko.pdf_utils.reader import PdfFileReader
 from pyhanko.pdf_utils.writer import copy_into_new_writer
 from pyhanko.sign import fields, signers, validation
@@ -49,10 +53,21 @@ try:
     import pkcs11  # lgtm [py/unused-import]
     pkcs11_available = True
 except ImportError:
+    pkcs11 = None
     pkcs11_available = False
 
+P11_PIN_ENV_VAR = "PYHANKO_PKCS11_PIN"
 
-def logging_setup(log_configs):
+
+class NoStackTraceFormatter(logging.Formatter):
+    def formatException(self, ei) -> str:
+        return ""
+
+
+LOG_FORMAT_STRING = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
+
+
+def logging_setup(log_configs, verbose: bool):
     log_config: LogConfig
     for module, log_config in log_configs.items():
         cur_logger = logging.getLogger(module)
@@ -62,11 +77,15 @@ def logging_setup(log_configs):
                 handler = logging.StreamHandler(sys.stdout)
             else:
                 handler = logging.StreamHandler()
+            # when logging to the console, don't output stack traces
+            # unless in verbose mode
+            if verbose:
+                formatter = logging.Formatter(LOG_FORMAT_STRING)
+            else:
+                formatter = NoStackTraceFormatter(LOG_FORMAT_STRING)
         else:
             handler = logging.FileHandler(log_config.output)
-        formatter = logging.Formatter(
-            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
-        )
+            formatter = logging.Formatter(LOG_FORMAT_STRING)
         handler.setFormatter(formatter)
         cur_logger.addHandler(handler)
 
@@ -78,18 +97,25 @@ def pyhanko_exception_manager():
         yield
     except click.ClickException:
         raise
+    except misc.PdfStrictReadError as e:
+        exception = e
+        msg = (
+            "Failed to read PDF file in strict mode; rerun with "
+            "--no-strict-syntax to try again.\n"
+            f"Error message: {e.msg}"
+        )
     except misc.PdfReadError as e:
         exception = e
-        msg = "Failed to read PDF file."
+        msg = f"Failed to read PDF file: {e.msg}"
     except misc.PdfWriteError as e:
         exception = e
-        msg = "Failed to write PDF file."
+        msg = f"Failed to write PDF file: {e.msg}"
     except SigningError as e:
         exception = e
-        msg = "Error raised while producing signed file."
+        msg = f"Error raised while producing signed file: {e.msg}"
     except LayoutError as e:
         exception = e
-        msg = "Error raised while producing signature layout."
+        msg = f"Error raised while producing signature layout: {e.msg}"
     except Exception as e:
         exception = e
         msg = "Generic processing error."
@@ -112,8 +138,18 @@ class Ctx(Enum):
     NEW_FIELD_SPEC = auto()
     PREFER_PSS = auto()
     DETACH_PEM = auto()
+    LENIENT = auto()
 
 
+def _warn_empty_passphrase():
+    click.echo(
+        click.style(
+            "WARNING: passphrase is empty. If you intended to sign with an "
+            "unencrypted private key, use --no-pass instead.",
+            bold=True
+        )
+    )
+
 @click.group()
 @click.version_option(prog_name='pyHanko', version=__version__)
 @click.option('--config',
@@ -173,7 +209,7 @@ def cli(ctx, config, verbose):
                 level=logging.WARNING, output=log_output
             )
 
-    logging_setup(log_config)
+    logging_setup(log_config, verbose)
 
     if verbose:
         logging.debug("Running with --verbose")
@@ -409,6 +445,14 @@ def _signature_status_str(status_callback, pretty_print, executive_summary):
             return 'INVALID', False
 
 
+def _attempt_iso_dt_parse(dt_str) -> datetime:
+    try:
+        dt = isoparse(dt_str)
+    except ValueError:
+        raise click.ClickException(f"datetime {dt_str!r} could not be parsed")
+    return dt
+
+
 # TODO add an option to do LTV, but guess the profile
 @signing.command(name='validate', help='validate signatures')
 @click.argument('infile', type=click.File('rb'))
@@ -439,6 +483,14 @@ def _signature_status_str(status_callback, pretty_print, executive_summary):
               help='Treat revocation info as retroactively valid '
                    '(i.e. ignore thisUpdate timestamp)',
               type=bool, is_flag=True, default=False, show_default=True)
+@click.option('--validation-time',
+              help=(
+                   'Override the validation time (ISO 8601 date). '
+                   'The special value \'claimed\' causes the validation time '
+                   'claimed by the signer to be used. Revocation checking '
+                   'will be disabled. Option ignored in LTV mode.'
+              ),
+              type=str, required=False)
 @click.option('--password', required=False, type=str,
               help='password to access the file (can also be read from stdin)')
 @click.option('--no-diff-analysis', default=False, type=bool, is_flag=True,
@@ -452,6 +504,7 @@ def _signature_status_str(status_callback, pretty_print, executive_summary):
 )
 @click.option('--no-strict-syntax',
               help='Attempt to ignore syntactical problems in the input file '
+                   'and enable signature validation in hybrid-reference files.'
                    '(warning: this may affect validation results in unexpected '
                    'ways.)',
               type=bool, is_flag=True, default=False, show_default=True)
@@ -461,7 +514,9 @@ def validate_signatures(ctx, infile, executive_summary,
                         other_certs, ltv_profile, force_revinfo,
                         soft_revocation_check, no_revocation_check, password,
                         retroactive_revinfo, detached, no_diff_analysis,
-                        no_strict_syntax):
+                        validation_time, no_strict_syntax):
+
+    no_revocation_check |= validation_time is not None
 
     if no_revocation_check:
         soft_revocation_check = True
@@ -472,6 +527,10 @@ def validate_signatures(ctx, infile, executive_summary,
         )
 
     if ltv_profile is not None:
+        if validation_time is not None:
+            raise click.ClickException(
+                "--validation-time is not compatible with --ltv-profile"
+            )
         ltv_profile = RevocationInfoValidationType(ltv_profile)
 
     vc_kwargs = _build_vc_kwargs(
@@ -480,6 +539,12 @@ def validate_signatures(ctx, infile, executive_summary,
         allow_fetching=False if no_revocation_check else None
     )
 
+    use_claimed_validation_time = False
+    if validation_time == 'claimed':
+        use_claimed_validation_time = True
+    elif validation_time is not None:
+        vc_kwargs['moment'] = _attempt_iso_dt_parse(validation_time)
+
     key_usage_settings = _get_key_usage_settings(ctx, validation_context)
     vc_kwargs = _prepare_vc(
         vc_kwargs, soft_revocation_check=soft_revocation_check,
@@ -524,6 +589,8 @@ def validate_signatures(ctx, infile, executive_summary,
         all_signatures_ok = True
         for ix, embedded_sig in enumerate(r.embedded_regular_signatures):
             fingerprint: str = embedded_sig.signer_cert.sha256.hex()
+            if use_claimed_validation_time:
+                vc_kwargs['moment'] = embedded_sig.self_reported_timestamp
             (status_str, signature_ok) = _signature_status_str(
                 status_callback=lambda: _signature_status(
                     ltv_profile=ltv_profile, force_revinfo=force_revinfo,
@@ -686,12 +753,17 @@ def ltv_fix(ctx, infile, field, timestamp_url, apply_lta_timestamp,
               help='Treat revocation info as retroactively valid '
                    '(i.e. ignore thisUpdate timestamp)',
               type=bool, is_flag=True, default=False, show_default=True)
+@click.option('--no-strict-syntax',
+              help='Attempt to ignore syntactical problems in the input file '
+                   'and enable signature creation in hybrid-reference files.'
+                   '(warning: such documents may behave in unexpected ways)',
+              type=bool, is_flag=True, default=False, show_default=True)
 @click.pass_context
 def addsig(ctx, field, name, reason, location, certify, existing_only,
            timestamp_url, use_pades, use_pades_lta, with_validation_info,
            validation_context, trust_replace, trust, other_certs,
            style_name, stamp_url, prefer_pss, retroactive_revinfo,
-           detach, detach_pem):
+           detach, detach_pem, no_strict_syntax):
     ctx.obj[Ctx.EXISTING_ONLY] = existing_only or field is None
     ctx.obj[Ctx.TIMESTAMP_URL] = timestamp_url
     ctx.obj[Ctx.PREFER_PSS] = prefer_pss
@@ -701,7 +773,13 @@ def addsig(ctx, field, name, reason, location, certify, existing_only,
         ctx.obj[Ctx.SIG_META] = None
         return  # everything else doesn't apply
 
-    if use_pades or use_pades_lta:
+    if use_pades_lta:
+        use_pades = with_validation_info = True
+        if not timestamp_url:
+            raise click.ClickException(
+                "--timestamp-url is required for --use-pades-lta"
+            )
+    if use_pades:
         subfilter = fields.SigSeedSubFilter.PADES
     else:
         subfilter = fields.SigSeedSubFilter.ADOBE_PKCS7_DETACHED
@@ -731,11 +809,12 @@ def addsig(ctx, field, name, reason, location, certify, existing_only,
     ctx.obj[Ctx.NEW_FIELD_SPEC] = new_field_spec
     ctx.obj[Ctx.STAMP_STYLE] = _select_style(ctx, style_name, stamp_url)
     ctx.obj[Ctx.STAMP_URL] = stamp_url
+    ctx.obj[Ctx.LENIENT] = no_strict_syntax
 
 
-def _open_for_signing(infile_path, signer_cert=None, signer_key=None):
+def _open_for_signing(infile_path, lenient, signer_cert=None, signer_key=None):
     infile = open(infile_path, 'rb')
-    writer = IncrementalPdfFileWriter(infile)
+    writer = IncrementalPdfFileWriter(infile, strict=not lenient)
 
     # TODO make this an option higher up the tree
     # TODO mention filename in prompt
@@ -814,7 +893,7 @@ async def async_detached_sig(signer: signers.Signer, infile_path, outfile,
 
 def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
                          timestamp_url, signature_meta, existing_fields_only,
-                         style, text_params, new_field_spec):
+                         style, text_params, new_field_spec, lenient):
     with pyhanko_exception_manager():
         if timestamp_url is not None:
             timestamper = HTTPTimeStamper(timestamp_url)
@@ -822,7 +901,7 @@ def addsig_simple_signer(signer: signers.SimpleSigner, infile_path, outfile,
             timestamper = None
         writer = _open_for_signing(
             infile_path, signer_cert=signer.signing_cert,
-            signer_key=signer.signing_key
+            signer_key=signer.signing_key, lenient=lenient
         )
 
         generic_sign_pdf(
@@ -880,8 +959,12 @@ def grab_certs(files):
 @click.option('--passfile', help='file containing the passphrase '
               'for the private key', required=False, type=click.File('r'),
               show_default='stdin')
+@click.option('--no-pass',
+              help='assume the private key file is unencrypted',
+              type=bool, is_flag=True, default=False, show_default=True)
 @click.pass_context
-def addsig_pemder(ctx, infile, outfile, key, cert, chain, pemder_setup, passfile):
+def addsig_pemder(ctx, infile, outfile, key, cert, chain, pemder_setup,
+                  passfile, no_pass):
     signature_meta = ctx.obj[Ctx.SIG_META]
     existing_fields_only = ctx.obj[Ctx.EXISTING_ONLY]
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
@@ -914,8 +997,10 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, pemder_setup, passfile
     elif passfile is not None:
         passphrase = passfile.readline().strip().encode('utf-8')
         passfile.close()
-    elif pemder_config.prompt_passphrase:
+    elif pemder_config.prompt_passphrase and not no_pass:
         passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
+        if not passphrase:
+            _warn_empty_passphrase()
     else:
         passphrase = None
 
@@ -930,7 +1015,8 @@ def addsig_pemder(ctx, infile, outfile, key, cert, chain, pemder_setup, passfile
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
         style=ctx.obj[Ctx.STAMP_STYLE], text_params=get_text_params(ctx),
-        new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC]
+        new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+        lenient=ctx.obj.get(Ctx.LENIENT, False)
     )
 
 
@@ -1003,7 +1089,8 @@ def addsig_pkcs12(ctx, infile, outfile, pfx, chain, passfile, p12_setup):
         signature_meta=signature_meta,
         existing_fields_only=existing_fields_only,
         style=ctx.obj[Ctx.STAMP_STYLE], text_params=get_text_params(ctx),
-        new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC]
+        new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+        lenient=ctx.obj.get(Ctx.LENIENT, False)
     )
 
 
@@ -1020,16 +1107,15 @@ def _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url):
         else:
             timestamper = None
 
-        with open(infile, 'rb') as inf:
-            generic_sign_pdf(
-                writer=IncrementalPdfFileWriter(inf),
-                outfile=outfile,
-                signature_meta=ctx.obj[Ctx.SIG_META], signer=signer,
-                timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
-                new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
-                existing_fields_only=ctx.obj[Ctx.EXISTING_ONLY],
-                text_params=get_text_params(ctx)
-            )
+        generic_sign_pdf(
+            writer=_open_for_signing(infile, ctx.obj.get(Ctx.LENIENT, False)),
+            outfile=outfile,
+            signature_meta=ctx.obj[Ctx.SIG_META], signer=signer,
+            timestamper=timestamper, style=ctx.obj[Ctx.STAMP_STYLE],
+            new_field_spec=ctx.obj[Ctx.NEW_FIELD_SPEC],
+            existing_fields_only=ctx.obj[Ctx.EXISTING_ONLY],
+            text_params=get_text_params(ctx)
+        )
 
 
 @click.argument('infile', type=readable_file)
@@ -1072,22 +1158,44 @@ def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
             logger.error(msg, exc_info=e)
             raise click.ClickException(msg)
     else:
-        if not (lib and token_label and cert_label):
+        if not (lib and cert_label):
             raise click.ClickException(
-                "The parameters --lib, --token-label and --cert-label "
-                "are required."
+                "The parameters --lib and --cert-label are required."
             )
+
+        pinentry_mode = (
+            PKCS11PinEntryMode.SKIP if skip_user_pin
+            else PKCS11PinEntryMode.PROMPT
+        )
+
         pkcs11_config = PKCS11SignatureConfig(
             module_path=lib, cert_label=cert_label, key_label=key_label,
-            slot_no=slot_no, token_label=token_label,
-            prompt_pin=not skip_user_pin, raw_mechanism=raw_mechanism
+            slot_no=slot_no, token_criteria=TokenCriteria(token_label),
+            # for now, DEFER requires a config file
+            prompt_pin=pinentry_mode,
+            raw_mechanism=raw_mechanism
         )
 
     pin = pkcs11_config.user_pin
-    if pin is None and pkcs11_config.prompt_pin:  # pragma: nocover
+
+    # try to fetch the PIN from an env var
+    if pin is None:
+        pin_env = os.environ.get(P11_PIN_ENV_VAR, None)
+        if pin_env:
+            pin = pin_env.strip()
+
+    if pkcs11_config.prompt_pin == PKCS11PinEntryMode.PROMPT \
+            and pin is None:  # pragma: nocover
         pin = getpass.getpass(prompt='PKCS#11 user PIN: ')
-    with pkcs11.PKCS11SigningContext(pkcs11_config, user_pin=pin) as signer:
-        _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
+    try:
+        with pkcs11.PKCS11SigningContext(pkcs11_config, user_pin=pin) as signer:
+            _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
+    except pkcs11.PKCS11Error as e:
+        # This will catch errors that happen during PKCS#11 setup
+        # TODO make sure PKCS11Signer doesn't leak PCKS11Errors in its
+        #  generic signing API
+        logger.error("PKCS#11 error", exc_info=e)
+        raise click.ClickException(f"PKCS#11 error: [{type(e).__name__}] {e}")
 
 
 @click.argument('infile', type=readable_file)
@@ -1112,7 +1220,12 @@ def addsig_beid(ctx, infile, outfile, lib, use_auth_cert, slot_no):
         lib = cli_config.beid_module_path
 
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
-    session = beid.open_beid_session(lib, slot_no=slot_no)
+
+    try:
+        session = beid.open_beid_session(lib, slot_no=slot_no)
+    except pkcs11.PKCS11Error as e:
+        logger.error("PKCS#11 error", exc_info=e)
+        raise click.ClickException(f"PKCS#11 error: [{type(e).__name__}] {e}")
     with session:
         signer = beid.BEIDSigner(session, use_auth_cert=use_auth_cert)
         _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
@@ -1344,13 +1457,21 @@ def decrypt_with_password(infile, outfile, password, force):
 @click.option('--passfile', required=False, type=click.File('rb'),
               help='file containing the passphrase for the private key',
               show_default='stdin')
+@click.option('--no-pass',
+              help='assume the private key file is unencrypted',
+              type=bool, is_flag=True, default=False, show_default=True)
 @decrypt_force_flag
-def decrypt_with_pemder(infile, outfile, key, cert, passfile, force):
-    if passfile is None:
-        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
-    else:
+def decrypt_with_pemder(infile, outfile, key, cert, passfile, force, no_pass):
+    if passfile is not None:
         passphrase = passfile.read()
         passfile.close()
+    elif not no_pass:
+        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
+        if not passphrase:
+            _warn_empty_passphrase()
+    else:
+        passphrase = None
+
     sedk = crypt.SimpleEnvelopeKeyDecrypter.load(
         key, cert, key_passphrase=passphrase
     )
