@@ -10,23 +10,25 @@ classes.
 
 import os
 from dataclasses import dataclass
+from datetime import datetime
 from enum import Enum
+from io import BytesIO
+from typing import Callable, Generator, Iterable, Optional, TypeVar
 
 __all__ = [
-    'PdfError', 'PdfReadError', 'PdfWriteError', 'PdfStreamError',
-    'IndirectObjectExpected',
-    'get_and_apply', 'get_courier', 'OrderedEnum',
+    'PdfError', 'PdfReadError', 'PdfStrictReadError',
+    'PdfWriteError', 'PdfStreamError', 'IndirectObjectExpected',
+    'get_and_apply', 'OrderedEnum', 'StringWithLanguage',
     'is_regular_character',
     'read_non_whitespace', 'read_until_whitespace', 'read_until_regex',
     'skip_over_whitespace', 'skip_over_comment', 'instance_test', 'peek',
     'assert_writable_and_random_access', 'prepare_rw_output_stream',
     'finalise_output',
     'DEFAULT_CHUNK_SIZE', 'chunked_write', 'chunked_digest', 'chunk_stream',
-    'ConsList', 'Singleton', 'rd'
+    'ConsList', 'Singleton', 'rd', 'isoparse'
 ]
 
-from io import BytesIO
-from typing import Callable, Generator, Iterable, TypeVar
+import pytz
 
 DEFAULT_CHUNK_SIZE = 4096
 """
@@ -86,41 +88,62 @@ def read_non_whitespace(stream, seek_back=False, allow_eof=False):
     Finds and reads the next non-whitespace character (ignores whitespace).
     """
     tok = PDF_WHITESPACE[0]
-    while tok in PDF_WHITESPACE:
-        if not tok:
-            if allow_eof:
-                return b''
-            else:
-                raise PdfStreamError('Stream ended prematurely')
-        tok = stream.read(1)
+    while True:
+        while tok in PDF_WHITESPACE:
+            if not tok:
+                if allow_eof:
+                    return b''
+                else:
+                    raise PdfStreamError('Stream ended prematurely')
+            tok = stream.read(1)
+        # Deal with comments
+        if tok != b'%':
+            break
+        else:
+            stream.seek(-1, os.SEEK_CUR)
+            skip_over_comment(stream)
+            tok = PDF_WHITESPACE[0]
     if seek_back:
         stream.seek(-1, os.SEEK_CUR)
     return tok
 
 
-def skip_over_whitespace(stream):
+def skip_over_whitespace(stream, stop_after_eol=False) -> bool:
     """
     Similar to readNonWhitespace, but returns a Boolean if more than
     one whitespace character was read.
 
     Will return the cursor to before the first non-whitespace character
-    encountered.
+    encountered, or after the first end-of-line sequence if one is encountered.
     """
     tok = PDF_WHITESPACE[0]
     cnt = 0
     while tok in PDF_WHITESPACE:
         tok = stream.read(1)
         cnt += 1
+        if stop_after_eol:
+            if tok == b'\n':
+                return cnt > 1
+            elif tok == b'\r':
+                # read the next char and check if it's a LF
+                if stream.read(1) == b'\n':
+                    return cnt > 1
+                # if not, break here; we need to seek back one position
+                # (CR by itself also counts as an EOL sequence)
+                break
+
     stream.seek(-1, os.SEEK_CUR)
     return cnt > 1
 
 
-def skip_over_comment(stream):
+def skip_over_comment(stream) -> bool:
     tok = stream.read(1)
     stream.seek(-1, 1)
     if tok == b'%':
         while tok not in (b'\n', b'\r'):
             tok = stream.read(1)
+        return True
+    return False
 
 
 def read_until_regex(stream, regex, ignore_eof=False):
@@ -149,18 +172,40 @@ def read_until_regex(stream, regex, ignore_eof=False):
     return name
 
 
+@dataclass(frozen=True)
+class StringWithLanguage:
+    """
+    A string with a language attached to it.
+    """
+
+    value: str
+    lang_code: Optional[str] = None
+    country_code: Optional[str] = None
+
+    def __str__(self):
+        return self.value
+
+
 class PdfError(Exception):
-    pass
+
+    def __init__(self, msg: str, *args):
+        self.msg = msg
+        super().__init__(msg, *args)
 
 
 class PdfReadError(PdfError):
     pass
 
 
-class IndirectObjectExpected(PdfReadError):
+class PdfStrictReadError(PdfReadError):
     pass
 
 
+class IndirectObjectExpected(PdfReadError):
+    def __init__(self, msg: Optional[str] = None):
+        super().__init__(msg=msg or "indirect object expected")
+
+
 class PdfWriteError(PdfError):
     pass
 
@@ -180,20 +225,6 @@ def peek(itr):
     return first, _itr()
 
 
-def get_courier():
-    """
-    :return:
-        A resource dictionary representing the standard Courier font
-        (or one of its metric equivalents).
-    """
-    from .generic import DictionaryObject, pdf_name
-    return DictionaryObject({
-        pdf_name('/Type'): pdf_name('/Font'),
-        pdf_name('/Subtype'): pdf_name('/Type1'),
-        pdf_name('/BaseFont'): pdf_name('/Courier')
-    })
-
-
 class OrderedEnum(Enum):
     """
     Ordered enum (from the Python documentation)
@@ -446,3 +477,18 @@ def finalise_output(orig_output, returned_output):
         return orig_output
     return returned_output
 
+
+def isoparse(dt_str: str) -> datetime:
+    try:
+        # Try to import the ISO parser from dateutil, if available
+        from dateutil.parser import isoparse as parse
+    except ImportError:  # pragma: nocover
+        # if not, call fromisoformat in the standard library
+        # (only implements a subset of ISO 8601)
+        parse = datetime.fromisoformat
+
+    dt = parse(dt_str)
+    if dt.tzinfo is None:
+        # assume UTC
+        dt = dt.replace(tzinfo=pytz.UTC)
+    return dt
