@@ -6,6 +6,7 @@ seamlessly plugged into a :class:`~.signers.PdfSigner`.
 import asyncio
 import binascii
 import logging
+import struct
 import warnings
 from dataclasses import dataclass
 from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
@@ -231,7 +232,7 @@ def select_pkcs11_signing_params(
 
     pre_sign_transform = None
     post_sign_transform = None
-    kwargs = {}
+    kwargs: Dict[str, Any] = {}
 
     try:
         signature_algo = signature_mechanism.signature_algo
@@ -286,6 +287,27 @@ def select_pkcs11_signing_params(
         kwargs['mechanism_param'] = (
             pss_digest_param, pss_mgf_param, pss_salt_len
         )
+    elif signature_algo == 'ed25519':
+        if use_raw_mechanism:
+            # Note: Ed25519-ph isn't the same thing.
+            raise NotImplementedError("Ed25519 not available in raw mode")
+        kwargs['mechanism'] = Mechanism.EDDSA
+    elif signature_algo == 'ed448':
+        if use_raw_mechanism:
+            # Note: Ed448-ph isn't the same thing.
+            raise NotImplementedError("Ed448 not available in raw mode")
+        kwargs['mechanism'] = Mechanism.EDDSA
+        # Definition of the param type:
+        # typedef struct CK_EDDSA_PARAMS {
+        #    CK_BBOOL     phFlag;
+        #    CK_ULONG     ulContextDataLen;
+        #    CK_BYTE_PTR  pContextData;
+        # }  CK_EDDSA_PARAMS;
+        # We use native size and alignment here on purpose
+
+        # NOTE: I _think_ this is correct, but it looks like SoftHSMv2
+        # doesn't really care about the params, so maybe I'm wrong.
+        kwargs['mechanism_param'] = struct.pack('@?LP', False, 0, 0)
     else:
         raise NotImplementedError(
             f"Signature algorithm '{signature_algo}' is not supported."
