@@ -2,6 +2,7 @@ import hashlib
 import logging
 from dataclasses import dataclass
 from dataclasses import field as data_field
+from typing import Optional
 
 from asn1crypto import crl as asn1_crl
 from asn1crypto import ocsp as asn1_ocsp
@@ -25,6 +26,7 @@ __all__ = [
     'collect_validation_info'
 ]
 
+from ...pdf_utils.crypt import SerialisedCredential
 from ...pdf_utils.reader import PdfFileReader
 
 logger = logging.getLogger(__name__)
@@ -314,20 +316,20 @@ class DocumentSecurityStore:
             raise NoDSSFoundError() from e
 
         cert_refs = {}
-        cert_ref_list = get_and_apply(dss_dict, '/Certs', list, default=())
+        cert_ref_list = get_and_apply(dss_dict, '/Certs', list, default=[])
         for cert_ref in cert_ref_list:
             cert_stream: generic.StreamObject = cert_ref.get_object()
             cert: Certificate = Certificate.load(cert_stream.data)
             cert_refs[cert.issuer_serial] = cert_ref
 
-        ocsp_refs = get_and_apply(dss_dict, '/OCSPs', list, default=())
+        ocsp_refs = get_and_apply(dss_dict, '/OCSPs', list, default=[])
         ocsps = []
         for ocsp_ref in ocsp_refs:
             ocsp_stream: generic.StreamObject = ocsp_ref.get_object()
             resp = asn1_ocsp.OCSPResponse.load(ocsp_stream.data)
             ocsps.append(resp)
 
-        crl_refs = get_and_apply(dss_dict, '/CRLs', list, default=())
+        crl_refs = get_and_apply(dss_dict, '/CRLs', list, default=[])
         crls = []
         for crl_ref in crl_refs:
             crl_stream: generic.StreamObject = crl_ref.get_object()
@@ -456,7 +458,8 @@ class DocumentSecurityStore:
     @classmethod
     def add_dss(cls, output_stream, sig_contents, *, certs=None,
                 ocsps=None, crls=None, paths=None, validation_context=None,
-                force_write: bool = False, embed_roots: bool = True):
+                force_write: bool = False, embed_roots: bool = True,
+                file_credential: Optional[SerialisedCredential] = None):
         """
         Wrapper around :meth:`supply_dss_in_writer`.
 
@@ -497,9 +500,14 @@ class DocumentSecurityStore:
 
             .. warning::
                 This only applies to paths, not the ``certs`` parameter.
+        :param file_credential:
+            .. versionadded:: 0.13.0
 
+            Serialised file credential, to update encrypted files.
         """
         pdf_out = IncrementalPdfFileWriter(output_stream)
+        if pdf_out.security_handler is not None and file_credential is not None:
+            pdf_out.security_handler.authenticate(file_credential)
         dss = cls.supply_dss_in_writer(
             pdf_out, sig_contents, certs=certs, ocsps=ocsps,
             crls=crls, paths=paths, validation_context=validation_context,
