@@ -11,8 +11,11 @@ from pyhanko_certvalidator.errors import PathBuildingError, PathValidationError
 from pyhanko_certvalidator.validate import ACValidationResult, async_validate_ac
 
 from pyhanko.sign.general import (
+    CMSExtractionError,
+    CMSStructuralError,
     MultivaluedAttributeError,
     NonexistentAttributeError,
+    SignedDataCerts,
     check_ess_certid,
     extract_certificate_info,
     extract_signer_info,
@@ -21,7 +24,8 @@ from pyhanko.sign.general import (
 )
 
 from ...pdf_utils import misc
-from .errors import SignatureValidationError, WeakHashAlgorithmError
+from ..ades.report import AdESFailure, AdESIndeterminate
+from . import errors
 from .settings import KeyUsageConstraints
 from .status import (
     CAdESSignerAttributeAssertions,
@@ -42,8 +46,8 @@ __all__ = [
     'collect_timing_info', 'validate_tst_signed_data',
     'async_validate_detached_cms', 'cms_basic_validation',
     'compute_signature_tst_digest', 'extract_tst_data',
-    'extract_self_reported_ts',
-    'collect_signer_attr_status',
+    'extract_self_reported_ts', 'extract_certs_for_validation',
+    'collect_signer_attr_status', 'validate_algorithm_protection'
 ]
 
 logger = logging.getLogger(__name__)
@@ -61,7 +65,7 @@ def _check_signing_certificate(cert: x509.Certificate,
         except NonexistentAttributeError:
             return None
         except MultivaluedAttributeError as e:
-            raise SignatureValidationError(
+            raise errors.SignatureValidationError(
                 "Wrong cardinality for signing certificate attribute"
             ) from e
 
@@ -71,12 +75,72 @@ def _check_signing_certificate(cert: x509.Certificate,
 
     if attr is None:
         # if neither attr is present -> no constraints
-        return True
+        return
 
     # we only care about the first value, the others limit the set of applicable
     # CA certs
     certid = attr['certs'][0]
-    return check_ess_certid(cert, certid)
+
+    if not check_ess_certid(cert, certid):
+        raise errors.SignatureValidationError(
+            f"Signing certificate attribute does not match selected "
+            f"signer's certificate for subject"
+            f"\"{cert.subject.human_friendly}\".",
+            ades_subindication=AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
+        )
+
+
+def validate_algorithm_protection(
+        attrs: cms.CMSAttributes,
+        claimed_digest_algorithm_obj: cms.DigestAlgorithm,
+        claimed_signature_algorithm_obj: cms.SignedDigestAlgorithm):
+    """
+    Internal API to validate the CMS algorithm protection attribute
+    defined in :rfc:`6211`, if present.
+
+    :param attrs:
+        A CMS attribute list.
+    :param claimed_digest_algorithm_obj:
+        The claimed (i.e. unprotected) digest algorithm value.
+    :param claimed_signature_algorithm_obj:
+        The claimed (i.e. unprotected) signature algorithm value.
+    :raises errors.CMSStructuralError:
+        if multiple CMS protection attributes are present
+    :raises errors.CMSAlgorithmProtectionError:
+        if a mismatch is detected
+    """
+
+    try:
+        cms_algid_protection = find_unique_cms_attribute(
+            attrs, 'cms_algorithm_protection'
+        )
+    except NonexistentAttributeError:
+        # TODO make this optional to enforce?
+        cms_algid_protection = None
+    except MultivaluedAttributeError:
+        raise CMSStructuralError(
+            "Multiple CMS algorithm protection attributes present",
+        )
+    if cms_algid_protection is not None:
+        auth_digest_algorithm = \
+            cms_algid_protection['digest_algorithm'].native
+        if auth_digest_algorithm != claimed_digest_algorithm_obj.native:
+            raise errors.CMSAlgorithmProtectionError(
+                "Digest algorithm does not match CMS algorithm protection "
+                "attribute.",
+            )
+        signed_sig_algorithm = \
+            cms_algid_protection['signature_algorithm'].native
+        if signed_sig_algorithm is None:
+            raise errors.CMSAlgorithmProtectionError(
+                "CMS algorithm protection attribute not valid for signed "
+                "data",
+            )
+        elif signed_sig_algorithm != claimed_signature_algorithm_obj.native:
+            raise errors.CMSAlgorithmProtectionError(
+                "Signature mechanism does not match CMS algorithm "
+                "protection attribute.",
+            )
 
 
 def validate_sig_integrity(signer_info: cms.SignerInfo,
@@ -119,7 +183,7 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
     digest_algorithm_obj = signer_info['digest_algorithm']
     md_algorithm = digest_algorithm_obj['algorithm'].native
     if md_algorithm in weak_hash_algorithms:
-        raise WeakHashAlgorithmError(md_algorithm)
+        raise errors.WeakHashAlgorithmError(md_algorithm)
     signature = signer_info['signature'].native
 
     # signed_attrs comes with some context-specific tagging.
@@ -134,58 +198,48 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
         prehashed = False
         # check the CMSAlgorithmProtection attr, if present
         try:
-            cms_algid_protection = find_unique_cms_attribute(
-                signed_attrs, 'cms_algorithm_protection'
+            validate_algorithm_protection(
+                signed_attrs,
+                claimed_digest_algorithm_obj=digest_algorithm_obj,
+                claimed_signature_algorithm_obj=signature_algorithm
             )
-        except NonexistentAttributeError:
-            cms_algid_protection = None
-        except MultivaluedAttributeError:
-            raise SignatureValidationError(
-                'Multiple CMS protection attributes present'
+        except CMSStructuralError as e:
+            raise errors.SignatureValidationError(
+                e.failure_message, ades_subindication=AdESFailure.FORMAT_FAILURE
+            )
+        except errors.CMSAlgorithmProtectionError as e:
+            raise errors.SignatureValidationError(
+                e.failure_message,
+                # these are conceptually failures, but AdES doesn't have
+                # them in its validation model, so 'GENERIC' it is.
+                #  (same applies to other such cases)
+                ades_subindication=AdESIndeterminate.GENERIC
             )
-        if cms_algid_protection is not None:
-            signed_digest_algorithm = \
-                cms_algid_protection['digest_algorithm'].native
-            if signed_digest_algorithm != digest_algorithm_obj.native:
-                raise SignatureValidationError(
-                    "Digest algorithm does not match CMS algorithm protection "
-                    "attribute."
-                )
-            signed_sig_algorithm = \
-                cms_algid_protection['signature_algorithm'].native
-            if signed_sig_algorithm is None:
-                raise SignatureValidationError(
-                    "CMS algorithm protection attribute not valid for signed "
-                    "data"
-                )
-            elif signed_sig_algorithm != signature_algorithm.native:
-                raise SignatureValidationError(
-                    "Signature mechanism does not match CMS algorithm "
-                    "protection attribute."
-                )
 
         # check the signing-certificate or signing-certificate-v2 attr
-        if not _check_signing_certificate(cert, signed_attrs):
-            raise SignatureValidationError(
-                f"Signing certificate attribute does not match selected "
-                f"signer's certificate for subject"
-                f"\"{cert.subject.human_friendly}\"."
-            )
+        # Note: Through the usual "full validation" call path, this check is
+        #   performed twice. AdES requires the check to be performed when
+        #   selecting the signer's certificate (which happens elsewhere), but
+        #   we keep this check for compatibility for those cases where
+        #   validate_sig_integrity is used standalone.
+        _check_signing_certificate(cert, signed_attrs)
 
         try:
             content_type = find_unique_cms_attribute(
                 signed_attrs, 'content_type'
             )
         except (NonexistentAttributeError, MultivaluedAttributeError):
-            raise SignatureValidationError(
+            raise errors.SignatureValidationError(
                 'Content type not found in signature, or multiple content-type '
-                'attributes present.'
+                'attributes present.',
+                ades_subindication=AdESFailure.FORMAT_FAILURE
             )
         content_type = content_type.native
         if content_type != expected_content_type:
-            raise SignatureValidationError(
+            raise errors.SignatureValidationError(
                 f'Content type {content_type} did not match expected value '
-                f'{expected_content_type}'
+                f'{expected_content_type}',
+                ades_subindication=AdESFailure.FORMAT_FAILURE
             )
 
         embedded_digest = extract_message_digest(signer_info)
@@ -208,6 +262,36 @@ def validate_sig_integrity(signer_info: cms.SignerInfo,
     return intact, valid
 
 
+def extract_certs_for_validation(signed_data: cms.SignedData) \
+        -> SignedDataCerts:
+    """
+    Extract certificates from a CMS signed data object for validation purposes,
+    identifying the signer's certificate in accordance with ETSI EN 319 102-1,
+    5.2.3.4.
+
+    :param signed_data:
+        The CMS payload.
+    :return:
+        The extracted certificates.
+    """
+
+    # TODO allow signer certificate to be obtained from elsewhere?
+
+    try:
+        cert_info = extract_certificate_info(signed_data)
+        cert = cert_info.signer_cert
+    except CMSExtractionError:
+        raise errors.SignatureValidationError(
+            'signer certificate not included in signature',
+            ades_subindication=AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
+        )
+    signer_info = extract_signer_info(signed_data)
+    signed_attrs = signer_info['signed_attrs']
+    # check the signing-certificate or signing-certificate-v2 attr
+    _check_signing_certificate(cert, signed_attrs)
+    return cert_info
+
+
 async def cms_basic_validation(
         signed_data: cms.SignedData,
         status_cls: Type[StatusType] = SignatureStatus,
@@ -224,7 +308,7 @@ async def cms_basic_validation(
     Internal API.
     """
     signer_info = extract_signer_info(signed_data)
-    cert_info = extract_certificate_info(signed_data)
+    cert_info = extract_certs_for_validation(signed_data)
     cert = cert_info.signer_cert
     other_certs = cert_info.other_certs
 
@@ -249,10 +333,15 @@ async def cms_basic_validation(
         raw_digest = md.finalize()
 
     # first, do the cryptographic identity checks
-    intact, valid = validate_sig_integrity(
-        signer_info, cert, expected_content_type=expected_content_type,
-        actual_digest=raw_digest, weak_hash_algorithms=weak_hash_algos
-    )
+    try:
+        intact, valid = validate_sig_integrity(
+            signer_info, cert, expected_content_type=expected_content_type,
+            actual_digest=raw_digest, weak_hash_algorithms=weak_hash_algos
+        )
+    except CMSStructuralError as e:
+        raise errors.SignatureValidationError(
+            "CMS structural error: " + e.failure_message
+        ) from e
 
     # if the data being encapsulated by the signature is itself invalid,
     #  this flag is set
@@ -260,23 +349,25 @@ async def cms_basic_validation(
     valid &= intact
 
     # next, validate trust
-    trusted = revoked = False
-    path = None
+    ades_status = path = None
     if valid:
-        validator = CertificateValidator(
-            cert, intermediate_certs=other_certs,
-            validation_context=validation_context
-        )
-        trusted, revoked, path = await status_cls.validate_cert_usage(
-            validator, key_usage_settings=key_usage_settings
-        )
+        try:
+            validator = CertificateValidator(
+                cert, intermediate_certs=other_certs,
+                validation_context=validation_context
+            )
+            ades_status, path = await status_cls.validate_cert_usage(
+                validator, key_usage_settings=key_usage_settings
+            )
+        except ValueError as e:
+            logger.error("Processing error in validation process", exc_info=e)
+            ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
 
     status_kwargs = status_kwargs or {}
     status_kwargs.update(
         intact=intact, valid=valid, signing_cert=cert,
         md_algorithm=md_algorithm, pkcs7_signature_mechanism=mechanism,
-        revoked=revoked, trusted=trusted,
-        validation_path=path
+        trust_problem_indic=ades_status, validation_path=path
     )
     return status_kwargs
 
@@ -474,7 +565,7 @@ async def validate_tst_signed_data(
     if isinstance(tst_info_bytes, core.ParsableOctetString):
         tst_info = tst_info_bytes.parsed
     if not isinstance(tst_info, tsp.TSTInfo):
-        raise SignatureValidationError(
+        raise errors.SignatureValidationError(
             "SignedData does not encapsulate TSTInfo"
         )
     # compare the expected TST digest against the message imprint
@@ -528,7 +619,7 @@ async def collect_signer_attr_status(
     except NonexistentAttributeError:
         signer_attrs = None
     except MultivaluedAttributeError as e:
-        raise SignatureValidationError(str(e)) from e
+        raise errors.SignatureValidationError(str(e)) from e
 
     result = {}
     cades_ac_results = None
@@ -538,6 +629,8 @@ async def collect_signer_attr_status(
         # process claimed attributes (no verification possible/required,
         # so this is independent of whether we have a validation context
         # available)
+        # TODO offer a strict mode where all attributes must be recognised
+        #  and/or at least parseable?
         claimed = ClaimedAttributes.from_iterable(
             claimed_asn1 if not isinstance(claimed_asn1, core.Void) else ()
         )
@@ -563,6 +656,8 @@ async def collect_signer_attr_status(
         # If we were able to validate AC's from the signers-attrs-v2 attribute,
         # compile the validation results
         if cades_ac_results is not None:
+            # TODO offer a strict mode where all attributes must be recognised
+            #  and/or at least parseable?
             certified = CertifiedAttributes.from_results(cades_ac_results)
         else:
             certified = None
