@@ -17,12 +17,13 @@ from pyhanko_certvalidator.path import ValidationPath
 from pyhanko_certvalidator.validate import ACValidationResult
 
 from ...pdf_utils.misc import OrderedEnum
+from ..ades.report import AdESFailure, AdESIndeterminate, AdESSubIndic
 from ..diff_analysis import (
     DiffResult,
     ModificationLevel,
     SuspiciousModification,
 )
-from .errors import SigSeedValueValidationError
+from .errors import SignatureValidationError, SigSeedValueValidationError
 from .settings import KeyUsageConstraints
 
 __all__ = [
@@ -57,18 +58,10 @@ class SignatureStatus:
     actually validates.
     """
 
-    trusted: bool
+    trust_problem_indic: Optional[AdESSubIndic]
     """
-    Reports whether the signer's certificate is trusted w.r.t. the currently
-    relevant validation context and key usage requirements.
-    """
-
-    # TODO add a separate expired flag
-
-    revoked: bool
-    """
-    Reports whether the signer's certificate has been revoked or not.
-    If this field is ``True``, then obviously :attr:`trusted` will be ``False``.
+    If not ``None``, provides the AdES subindication indication what went
+    wrong when validating the signer's certificate.
     """
 
     signing_cert: x509.Certificate
@@ -118,6 +111,27 @@ class SignatureStatus:
             cert_status = 'UNTRUSTED'
         yield cert_status
 
+    @property
+    def revoked(self) -> bool:
+        """
+        Reports whether the signer's certificate has been revoked or not.
+        If this field is ``True``, then obviously :attr:`trusted` will be
+        ``False``.
+        """
+        return self.trust_problem_indic in (
+            AdESFailure.REVOKED,
+            AdESIndeterminate.REVOKED_CA_NO_POE,
+            AdESIndeterminate.REVOKED_NO_POE
+        )
+
+    @property
+    def trusted(self) -> bool:
+        """
+        Reports whether the signer's certificate is trusted w.r.t. the currently
+        relevant validation context and key usage requirements.
+        """
+        return self.valid and self.trust_problem_indic is None
+
     # TODO explain in more detail.
     def summary(self, delimiter=','):
         """
@@ -145,25 +159,35 @@ class SignatureStatus:
         )
         cert: x509.Certificate = validator._certificate
 
-        revoked = trusted = False
         path = None
-
         try:
             # validate usage without going through pyhanko_certvalidator
             key_usage_settings.validate(cert)
             path = await validator.async_validate_usage(key_usage=set())
-            trusted = True
+            ades_status = None
         except InvalidCertificateError as e:
-            # TODO accumulate these somewhere
+            # TODO accumulate these somewhere?
             logger.warning(e)
+            ades_status = AdESIndeterminate.CHAIN_CONSTRAINTS_FAILURE
         except RevokedError:
-            revoked = True
-        except (PathValidationError, PathBuildingError) as e:
+            # TODO have certvalidator report with which of the three
+            #  revocation cases under AdES we're dealing with.
+            # (For now, assume the worst.)
+            ades_status = AdESFailure.REVOKED
+        except PathBuildingError as e:
             logger.warning(e)
-        if not trusted:
+            ades_status = AdESIndeterminate.NO_CERTIFICATE_CHAIN_FOUND
+        except PathValidationError as e:
+            logger.warning(e)
+            # TODO make error reporting in certvalidator more granular
+            #  so we can actually set proper AdES values
+            #  (e.g. expiration-related ones, distinguish between CA and EE,
+            #   revinfo freshness...)
+            ades_status = AdESIndeterminate.CERTIFICATE_CHAIN_GENERAL_FAILURE
+        if ades_status is not None:
             subj = cert.subject.human_friendly
             logger.warning(f"Chain of trust validation for {subj} failed.")
-        return trusted, revoked, path
+        return ades_status, path
 
     @property
     def _trust_anchor(self):
@@ -242,21 +266,45 @@ class CertifiedAttributeInfo(X509AttributeInfo):
     """
 
 
+def _handle_attr_err(attr_type: Optional[str], attr_kind: str,
+                     err: ValueError, fatal: bool):
+    attr_type_str = "unknown type" if not attr_type else f"type '{attr_type}'"
+    msg = (
+        f"Failed to parse {attr_kind} of {attr_type_str}: {err.args[0]}"
+    )
+    if fatal:
+        raise SignatureValidationError(
+            msg, ades_subindication=AdESFailure.FORMAT_FAILURE
+        ) from err
+    else:
+        logger.warning(msg, exc_info=err)
+
+
 class CertifiedAttributes:
     """
     Container class for extracted attribute certificate information.
     """
 
     @classmethod
-    def from_results(cls, results: Iterable[ACValidationResult]):
+    def from_results(cls, results: Iterable[ACValidationResult],
+                     parse_error_fatal=False):
         # first, classify the attributes and results by type
         by_type = defaultdict(lambda: ([], []))
         for result in results:
             for attr_type, attr in result.approved_attributes.items():
+                attr_type = None
+                try:
+                    attr_type = attr['type'].native
+                    values = list(attr['values'])  # force a surface-level parse
+                except ValueError as e:
+                    _handle_attr_err(
+                        attr_type, "certified attribute", e,
+                        fatal=parse_error_fatal
+                    )
+                    continue
                 type_values, type_results = by_type[attr_type]
-                type_values.extend(attr['values'])
+                type_values.extend(values)
                 type_results.append(result)
-
         # then, for each type, we package 'em up in a CertifiedAttributeInfo
         infos = CertifiedAttributes()
         for attr_type, (type_values, type_results) in by_type.items():
@@ -294,18 +342,28 @@ class ClaimedAttributes:
     """
 
     @classmethod
-    def from_iterable(cls, attrs: Iterable[cms.AttCertAttribute]):
+    def from_iterable(cls, attrs: Iterable[cms.AttCertAttribute],
+                      parse_error_fatal=False):
         infos = ClaimedAttributes()
-        by_type = defaultdict(lambda: ([], []))
+        by_type = defaultdict(list)
         for attr in attrs:
-            type_values, type_results = by_type[attr['type'].native]
-            type_values.extend(attr['values'])
-        for attr_type, (type_values, type_results) in by_type.items():
-            infos._attrs[attr_type] = CertifiedAttributeInfo(
+            attr_type = None
+            try:
+                attr_type = attr['type'].native
+                values = list(attr['values'])  # force a surface-level parse
+            except ValueError as e:
+                _handle_attr_err(
+                    attr_type, "claimed attribute", e,
+                    fatal=parse_error_fatal
+                )
+                continue
+            by_type[attr_type].extend(values)
+
+        for attr_type, type_values in by_type.items():
+            infos._attrs[attr_type] = X509AttributeInfo(
                 attr_type=cms.AttCertAttributeType(attr_type),
                 # (shallow) immutability
                 attr_values=tuple(type_values),
-                validation_results=tuple(type_results)
             )
         return infos
 
